#pragma once

#include <ArmarXCore/core/Component.h>

#include <RobotAPI/interface/units/HandUnitInterface.h>
#include <RobotAPI/interface/units/KinematicUnitInterface.h>
#include <RobotAPI/libraries/RobotAPIComponentPlugins/RobotUnitComponentPlugin.h>
#include <RobotAPI/libraries/armem/client/plugins/PluginUser.h>
#include <RobotAPI/libraries/armem/client/plugins/ReaderWriterPlugin.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>
#include <RobotAPI/libraries/skills/provider/SkillProviderComponentPlugin.h>
#include <RobotAPI/libraries/skills/provider/blueprints/SkillWithContextBlueprint.h>

#include <armarx/control/components/control_skill_provider/ComponentInterface.h>
#include <armarx/control/skills/skills/ZeroTorque.h>

namespace armarx::control::components::control_skill_provider
{
    class Component :
        virtual public ::armarx::Component,
        virtual public ::armarx::control::components::control_skill_provider::ComponentInterface,
        virtual public ::armarx::SkillProviderComponentPluginUser,
        virtual public ::armarx::armem::client::PluginUser
    {

    public:
        Component();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        /// Get the component's default name.
        static std::string GetDefaultName();


    protected:
        /// @see PropertyUser::createPropertyDefinitions()
        ::armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        /// @see armarx::ManagedIceObject::onInitComponent()
        void onInitComponent() override;

        /// @see armarx::ManagedIceObject::onConnectComponent()
        void onConnectComponent() override;

        /// @see armarx::ManagedIceObject::onDisconnectComponent()
        void onDisconnectComponent() override;

        /// @see armarx::ManagedIceObject::onExitComponent()
        void onExitComponent() override;


    private:
        struct Remote
        {
            ::armarx::KinematicUnitInterfacePrx kinematicUnit;
            ::armarx::HandUnitInterfacePrx leftHandUnit;
            ::armarx::HandUnitInterfacePrx rightHandUnit;
            std::experimental::observer_ptr<armarx::plugins::RobotUnitComponentPlugin>
                robotUnitPlugin;
            std::experimental::observer_ptr<
                armem::client::plugins::ReaderWriterPlugin<armem::robot_state::VirtualRobotReader>>
                robotReaderPlugin;
        } remote;

        struct SkillBlueprints
        {
            std::experimental::observer_ptr<
                armarx::skills::SkillWithContextBlueprint<skills::skills::ZeroTorque>>
                zeroTorque;
        } skillBlueprints;

        struct Properties
        {
            std::string robotName;
        } properties;

        static const std::string defaultName;
    };

} // namespace armarx::control::components::control_skill_provider
