#pragma once

#include <RobotAPI/libraries/skills/provider/SpecializedSkill.h>

#include <armarx/control/skills/aron/ShapeHandParams.aron.generated.h>
#include <armarx/control/skills/constants/constants.h>

#include "HandUnitServices.h"

namespace armarx::control::skills::skills
{

    class ShapeHand : public ::armarx::skills::SpecializedSkill<arondto::ShapeHandParams>
    {

    public:
        static ::armarx::skills::SkillDescription
        GetSkillDescription()
        {
            return ::armarx::skills::SkillDescription{
                .skillId = {.skillName = constants::SHAPE_HAND_SKILL_NAME},
                .description =
                    "Shapes the hand according to a predefined shape or passed joint angles",
                .timeout = ::armarx::Duration::Minutes(1),
                .parametersType = ParamType::ToAronType()};
        }

        ShapeHand(const HandUnitServices&);

    private:
        ::armarx::skills::Skill::MainResult main() override;
        Skill::MainResult waitUntilFinished(const NameValueMap& targetAngles);

    private:
        HandUnitServices srv_;

        float getAccuracy(const std::string& jointName);
    };

    class OpenHand : public ::armarx::skills::SpecializedSkill<arondto::OpenHandParams>
    {
    public:
        static ::armarx::skills::SkillDescription
        GetSkillDescription()
        {
            return ::armarx::skills::SkillDescription{
                .skillId = {.skillName = constants::OPEN_HAND_SKILL_NAME},
                .description = "Sets hand shape to 'Open'",
                .timeout = ::armarx::Duration::Minutes(1),
                .parametersType = ParamType::ToAronType()};
        }

        OpenHand();

    private:
        ::armarx::skills::Skill::MainResult main() override;
    };

    class CloseHand : public ::armarx::skills::SpecializedSkill<arondto::CloseHandParams>
    {
    public:
        static ::armarx::skills::SkillDescription
        GetSkillDescription()
        {
            return ::armarx::skills::SkillDescription{
                .skillId = {.skillName = constants::CLOSE_HAND_SKILL_NAME},
                .description = "Sets hand shape to 'Close'",
                .timeout = ::armarx::Duration::Minutes(1),
                .parametersType = ParamType::ToAronType()};
        }

        CloseHand();

    private:
        ::armarx::skills::Skill::MainResult main() override;
    };

} // namespace armarx::control::skills::skills
